@@ -128,8 +128,7 @@ def do_connect(env, args):
 
     # set explicit route to gateway
     gwr = providers.route.get_route(env.gateway)
-    providers.route.replace_route(
-        env.gateway, **{k: gwr.get(k) for k in ('via', 'dev', 'src', 'mtu')})
+    providers.route.replace_route(env.gateway, **gwr)
 
     # drop incoming traffic from VPN
     if not args.incoming:
@@ -172,7 +171,7 @@ def do_connect(env, args):
     exc_subnets = [(dest, providers.route.get_route(dest)) for dest in args.exc_subnets]
 
     # set up routes to the DNS and Windows name servers, subnets, and local aliases
-    ns = env.dns + (env.nbns if args.nbns else [])
+    ns = env.dns + env.dns6 + (env.nbns if args.nbns else [])
     for dest in chain(ns, args.subnets, args.aliases):
         providers.route.replace_route(dest, dev=env.tundev)
     else:
@@ -182,11 +181,11 @@ def do_connect(env, args):
 
     # restore routes to excluded subnets
     for dest, exc_route in exc_subnets:
-        providers.route.replace_route(dest, exc_route)
+        providers.route.replace_route(dest, **exc_route)
     else:
         providers.route.flush_cache()
         if args.verbose:
-            print("Restored routes for %d excluded subnets." % len(exc_subnets), file=stderr)
+            print("Restored routes for %d excluded subnets." % len(exc_subnets), exc_subnets, file=stderr)
 
 def do_post_connect(env, args):
     global providers
@@ -196,7 +195,7 @@ def do_post_connect(env, args):
     host_map = []
 
     if args.ns_hosts:
-        ns_names = [ (ip, ('dns%d.%s' % (ii, args.name),)) for ii, ip in enumerate(env.dns) ]
+        ns_names = [ (ip, ('dns%d.%s' % (ii, args.name),)) for ii, ip in enumerate(env.dns + env.dns6) ]
         if args.nbns:
             ns_names += [ (ip, ('nbns%d.%s' % (ii, args.name),)) for ii, ip in enumerate(env.nbns) ]
         host_map += ns_names
@@ -207,7 +206,7 @@ def do_post_connect(env, args):
 
     if args.verbose:
         print("Looking up %d hosts using VPN DNS servers..." % len(args.hosts), file=stderr)
-    providers.dns.configure(dns_servers=env.dns, search_domains=args.domain, bind_addresses=env.myaddrs)
+    providers.dns.configure(dns_servers=(env.dns + env.dns6), search_domains=args.domain, bind_addresses=env.myaddrs)
     for host in args.hosts:
         try:
            ips = providers.dns.lookup_host(host)
@@ -243,7 +242,7 @@ def do_post_connect(env, args):
     # run DNS queries in background to prevent idle timeout
     if args.prevent_idle_timeout:
         dev = env.tundev
-        dns = env.dns
+        dns = (env.dns + env.dns6)
         idle_timeout = env.idle_timeout
         setproctitle('vpn-slice --prevent-idle-timeout --name %s' % args.name)
         if args.verbose:
@@ -263,7 +262,7 @@ def do_post_connect(env, args):
             # pick random host or IP to look up without leaking any new information
             # about what we do/don't access within the VPN
             pool = args.hosts
-            pool += map(str, chain(env.dns, env.nbns, ((r.network_address) for r in args.subnets if r.prefixlen==r.max_prefixlen)))
+            pool += map(str, chain(env.dns, env.dns6, env.nbns, ((r.network_address) for r in args.subnets if r.prefixlen==r.max_prefixlen)))
             dummy = choice(pool)
             shuffle(dns)
             if args.verbose > 1:
@@ -325,12 +324,18 @@ def parse_env(environ=os.environ):
                 ad=orig_netaddr, nm=env.netmask, nml=env.netmasklen, nmi=env.network.netmask))
         assert env.network.netmask==env.netmask
 
-    # IPv6 network is determined by the netmask only
-    # (e.g. /16 supplied as part of the address, or ffff:ffff:ffff:ffff:: supplied as separate netmask)
-    if env.myaddr6:
-        env.network6 = env.netmask6.network if env.netmask6 else env.myaddr6.network
-        env.myaddr6 = env.myaddr6.ip
+    # Need to match behavior of original vpnc-script here
+    # Examples:
+    #   1) INTERNAL_IP6_ADDRESS=fe80::1, INTERNAL_IP6_NETMASK=fe80::/64  => interface of fe80::1/64,  network of fe80::/64
+    #   2) INTERNAL_IP6_ADDRESS=unset,   INTERNAL_IP6_NETMASK=fe80::1/64 => interface of fe80::1/64,  network of fe80::/64
+    #   3) INTERNAL_IP6_ADDRESS=2000::1, INTERNAL_IP6_NETMASK=unset      => interface of 2000::1/128, network of 2000::1/128
+    if env.myaddr6 or env.netmask6:
+        if not env.netmask6:
+            env.netmask6 = IPv6Network(env.myaddr6) # case 3 above, /128
+        env.myaddr6 = IPv6Interface(env.netmask6)
+        env.network6 = env.myaddr6.network
     else:
+        env.myaddr6 = None
         env.network6 = None
 
     env.myaddrs = list(filter(None, (env.myaddr, env.myaddr6)))
@@ -351,6 +356,15 @@ def parse_env(environ=os.environ):
                 pfx=pfx, n=n, ad=ad, nm=nm, nml=nml, nmi=net.netmask))
         env['split'+pfx.lower()].append(net)
 
+    for pfx, n in chain((('INC', n) for n in range(env.nsplitinc6)),
+                        (('EXC', n) for n in range(env.nsplitexc6))):
+        ad = IPv6Address(environ['CISCO_IPV6_SPLIT_%s_%d_ADDR' % (pfx, n)])
+        nml = int(environ['CISCO_IPV6_SPLIT_%s_%d_MASKLEN' % (pfx, n)])
+        net = IPv6Network(ad).supernet(new_prefix=nml)
+        if net.network_address != ad:
+            print("WARNING: IPv6 split network (CISCO_IPV6_SPLIT_%s_%d_{ADDR,MASKLEN}) %s/%d has host bits set, replacing with %s" % (pfx, n, ad, nml, net), file=stderr)
+        env['split'+pfx.lower()].append(net)
+
     return env
 
 # Parse command-line arguments and environment
@@ -374,6 +388,7 @@ def parse_args_and_env(args=None, environ=os.environ):
     g.add_argument('--no-ns-hosts', action='store_false', dest='ns_hosts', default=True, help='Do not add nameserver aliases to /etc/hosts (default is to name them dns0.tun0, etc.)')
     g.add_argument('--nbns', action='store_true', dest='nbns', help='Include NBNS (Windows/NetBIOS nameservers) as well as DNS nameservers')
     g = p.add_argument_group('Debugging options')
+    g.add_argument('--self-test', action='store_true', help='Stop after verifying that environment variables and providers are configured properly.')
     g.add_argument('-v','--verbose', default=0, action='count', help="Explain what %(prog)s is doing")
     p.add_argument('-V','--version', action='version', version='%(prog)s ' + __version__)
     g.add_argument('-D','--dump', action='store_true', help='Dump environment variables passed by caller')
@@ -423,42 +438,54 @@ def finalize_args_and_env(args, env):
 
 def main(args=None, environ=os.environ):
     global providers
-    p, args, env = parse_args_and_env(args, environ)
 
-    providers = slurpy()
-    for pn, pv in get_default_providers().items():
-        try:
-            if isinstance(pv, Exception):
-                raise pv
-            providers[pn] = pv()
-        except Exception as e:
-            print("WARNING: Couldn't configure {} provider: {}".format(pn, e), file=stderr)
-    missing_required = {p for p in ('route', 'process', 'hosts', 'dns') if p not in providers}
-    if missing_required:
-        raise SystemExit("Aborting because providers for %s are required; use --help for more information" % ' '.join(missing_required))
+    try:
+        p, args, env = parse_args_and_env(args, environ)
 
-    finalize_args_and_env(args, env)
+        providers = slurpy()
+        for pn, pv in get_default_providers().items():
+            try:
+                if isinstance(pv, Exception):
+                    raise pv
+                providers[pn] = pv()
+            except Exception as e:
+                print("WARNING: Couldn't configure {} provider: {}".format(pn, e), file=stderr)
+        missing_required = {p for p in ('route', 'process', 'hosts', 'dns') if p not in providers}
+        if missing_required:
+            raise RuntimeError("Aborting because providers for %s are required; use --help for more information" % ' '.join(missing_required))
 
-    if env.myaddr6 or env.netmask6:
-        print('WARNING: IPv6 address or netmask set, but this version of %s has only rudimentary support for them.' % p.prog, file=stderr)
-    if env.dns6:
-        print('WARNING: IPv6 DNS servers set, but this version of %s does not know how to handle them' % p.prog, file=stderr)
-    if any(v.startswith('CISCO_IPV6_SPLIT_') for v in environ):
-        print('WARNING: CISCO_IPV6_SPLIT_* environment variables set, but this version of %s does not handle them' % p.prog, file=stderr)
-    if args.dump:
-        exe = providers.process.pid2exe(args.ppid)
-        caller = '%s (PID %d)'%(exe, args.ppid) if exe else 'PID %d' % args.ppid
+        finalize_args_and_env(args, env)
 
-        print('Called by %s with environment variables for vpnc-script:' % caller, file=stderr)
-        width = max((len(envar) for var, envar, *rest in vpncenv if envar in environ), default=0)
-        for var, envar, *rest in vpncenv:
-            if envar in environ:
-                pyvar = var+'='+repr(env[var]) if var else 'IGNORED'
-                print('  %-*s => %s' % (width, envar, pyvar), file=stderr)
-        if env.splitinc:
-            print('  %-*s => %s=%r' % (width, 'CISCO_SPLIT_INC_*', 'splitinc', env.splitinc), file=stderr)
-        if env.splitexc:
-            print('  %-*s => %s=%r' % (width, 'CISCO_SPLIT_EXC_*', 'splitexc', env.splitexc), file=stderr)
+        if env.myaddr6 or env.netmask6:
+            print('WARNING: IPv6 address or netmask set. Support for IPv6 in %s should be considered BETA-QUALITY.' % p.prog, file=stderr)
+        if args.dump:
+            exe = providers.process.pid2exe(args.ppid)
+            caller = '%s (PID %d)'%(exe, args.ppid) if exe else 'PID %d' % args.ppid
+
+            print('Called by %s with environment variables for vpnc-script:' % caller, file=stderr)
+            width = max((len(envar) for var, envar, *rest in vpncenv if envar in environ), default=0)
+            for var, envar, *rest in vpncenv:
+                if envar in environ:
+                    pyvar = var+'='+repr(env[var]) if var else 'IGNORED'
+                    print('  %-*s => %s' % (width, envar, pyvar), file=stderr)
+            if env.splitinc:
+                print('  %-*s => %s=%r' % (width, 'CISCO_*SPLIT_INC_*', 'splitinc', env.splitinc), file=stderr)
+            if env.splitexc:
+                print('  %-*s => %s=%r' % (width, 'CISCO_*SPLIT_EXC_*', 'splitexc', env.splitexc), file=stderr)
+
+    except Exception as e:
+        if args.self_test:
+            print('******************************************************************************************', file=stderr)
+            print('*** Self-test did not pass. Double-check that you are running as root (e.g. with sudo) ***', file=stderr)
+            print('******************************************************************************************', file=stderr)
+        raise SystemExit(*e.args)
+
+    else:
+        if args.self_test:
+            print('***************************************************************************', file=stderr)
+            print('*** Self-test passed. Try using vpn-slice with openconnect or vpnc now. ***', file=stderr)
+            print('***************************************************************************', file=stderr)
+            raise SystemExit()
 
     if env.reason is None:
         raise SystemExit("Must be called as vpnc-script, with $reason set; use --help for more information")
